(* Nazev balicku : CAGD` *)

(* Popis : 
    Funkce pro vypocet parametrizaci a zobrazeni krivek a ploch geometrickeho \
modelovani, transformace techto objektu *)

(* Autor : Bohumir Bastl, ZCU, Plzen *)

BeginPackage["CAGD`", {"Graphics`PlotField`"}]

NormalizeVectors::usage = \
"Pro zadane vektory vrati jejich normalizovane (jednotkove) verze. Pouziti: NormalizeVectors[list vektoru tvaru {{_,_}..}]";\


ChordalParametrization::usage = \
"Pro zadane body v rovine vrati odpovidajici vektor chordalove parametrizace (parametrizace zohlednujici vzdalenosti bodu). Pouziti: ChordalParametrization[body]";\


Ferguson::usage = \
"Pro zadane vstupni hodnoty, tzn. dva body v rovine a tecne vektory v techto bodech, vypocte Fergusnovu kubiku. Pouziti: Ferguson[pocatecni bod,koncovy bod,pocatecni tecny vektor,koncovy tecny vektor,hodnoty parametru v pocatecnim a koncovem bode,promenna parametrizace]";\


CubicSpline::usage = "Funkce pro vypocet kubicke spline krivky pro zadane body a okrajove podminky. Funkce vraci parametrizace jednotlivych oblouku spline krivky ve tvaru {x(prom),y(prom),{prom,prom_min,prom_max}}. Pouziti:
    Parametrizations = CubicSpline[body,promenna parametrizace,Options]. Prvnim parametrem je tedy operne body splinu, druhym parametrem zvolena promenna vyslednych parametrizaci oblouku,poslednim parametrem jsou Options, ktere mohou obsahovat urceni typu parametrizace pomoci Parametrization->{Chordal,Classic} (Chordal znaci chordalovou parametrizaci, Classic predstavuje uniformni parametrizaci), urceni okrajovych podminek opmoci BoundaryConditions->{Clamped,SecondDerivatives,Periodic,Natural} (Clamped ... zadane tecne vektory na zacatku a na konci, SecondDerivatives ... zadane vektory druhych derivaci na zacatku a na konci, Periodic ... uzavreny spline, Natural ... prirozeny spline, tedy na zacatku a na konci jsou nulove druhe derivace) a take definici vektoru urcujicich okrajove podminky (nutne pouze v pripade, ze jsou vybrany okrajove podminky typu 'Clamped' nebo 'SecondDerivatives', kdy tento parametr obsahuje prave zadane okrajove podminky ve tvaru {vektor1,vektor2}, jinak neni potreba uvadet). Pokud nejsou zadany zadne 'Options' bude vypocten uniformni prirozeny kubicky spline.";\


PlotCubicSpline::usage = \
"Vykresleni kubickeho splinu, funkce vraci obrazek splinu, parametry jsou obdobne jako u CubicSpline.";\


Parametrization::usage = \
"Option to CubicSpline. It specifies the type of used parametrization - Chordal or Classic.";\


Color::usage = "Option to CubicSpline and QuadraticSpline and all plot functions for curves. It specifies the color of the output curve in HUE format."
\


BoundaryConditions::usage = \
"Opiton to CubicSpline. It specifies the type of used boundary conditions - Clamped, Periodic, SecondDerivatives, Natural.";\


Chordal::usage = "Option to Parametrization.";

Classic::usage = "Option to Patametrization.";

Clamped::usage = "Option to BoundaryConditions.";

Periodic::usage = "Option to BoundaryConditions.";

SecondDerivatives::usage = "Option to BoundaryConditions.";

Natural::usage = "Option to BoundaryConditions.";

QuadraticSpline::usage = \
"Funkce pro vypocet kvadraticke spline krivky pro zadane body a okrajovou podminku. Funkce vraci parametrizace jednotlivych oblouku spline krivky ve tvaru {x(prom),y(prom),{prom,prom_min,prom_max}}. Pouziti: Parametrizations = QuadraticSpline[body,okrajova podminka,Parametrization->{Chordal,Classic},promenna parametrizace]. Prvnim parametrem jsou tedy operne body splinu, druhym parametrem je tecny vektor v pocatecnim bode, tretim parametrem je zvoleny zpusob parametrizace splinu (Chordal znaci chordalovou parametrizaci, Classic predstavuje uniformni parametrizaci) a poslednim parametrem je zvolena promenna vyslednych parametrizaci oblouku splinu.";\


PlotQuadraticSpline::usage = \
"Vykresleni kvadratickeho splinu, funkce vraci obrazek splinu, parametry jsou obdobne jako u QuadraticSpline.";\


Bernstein::usage = \
"Vypocte i-ty Bernsteinuv polynom stupne n (polynom bude v promenne v). Pouziti: Bernstein[i,n,promenna]";\


BezierCurve::usage = \
"Vypocte parametrizaci rovinne (body ridici polygonu maji dve souradnice) i prostorove (body ridici polygonu maji tri souradnice) Bezierovy krivky pro zadany ridici polygon. Symbol parametru se zadava jako posledni parametr funkce. Pouziti: BezierCurve[{body ridiciho polygonu},promenna parametrizace]";\


PlotBezierCurve::usage = \
"Vykresleni rovinne i prostorove Bezierovy krivky zadane ridicim polygonem. Parametry stejne jako u BezierCurve, navic je pouze posledni parametr urcujici barvu krivku (cislo v rozsahu 0, 1). Pouziti: PlotBezierCurve[{body ridiciho polygonu},promenna parametrizace,barva]";\


RatBezierCurve::usage = \
"Vypocte parametrizaci rovinne (body ridici polygonu maji tri souradnice, posledni souradnice je vaha bodu) i prostorove (body ridici polygonu maji ctyri souradnice,  posledni souradnice je vaha bodu) racionalni Bezierovy krivky pro zadany ridici polygon. Symbol parametru se zadava jako posledni parametr funkce. Pouziti: RatBezierCurve[{body ridiciho polygonu},promenna parametrizace]";\


PlotRatBezierCurve::usage = \
"Vykresleni rovinne i prostorove racionalni Bezierovy krivky zadane ridicim polygonem. Parametry stejne jako u RatBezierCurve, navic je pouze posledni parametr urcujici barvu krivku (cislo v rozsahu 0, 1). Pouziti: PlotRatBezierCurve[{body ridiciho polygonu},promenna parametrizace,barva]";\


PlotRatBezierCurveClassic::usage = \
"Podobne jako PlotRatBezierCurve, pouze se jinak zobrazuje ridici polygon v pripade, ze obsahuje ridici body s nulovou vahou.";\


BezierSurface::usage = 
    "Vypocte parametrizaci Bezierovy plochy pro zadanou ridici sit. Ridici sit se zadava jako trojrozmerne pole typu {{{_,_,_}..}..}, ktere predstavuje matici bodu ridici site {{P00,...,P0n},...,{Pm0,...,Pmn}} a kazde Pij je bod v prostoru. Druhym parametrem je vektor o dvou slozkach, obsahujici symboly pouzite pro parametry vysledne parametrizace. Pouziti: BezierSurface[{ridici sit},{parametr1,parametr2}]";\


PlotBezierSurface::usage = \
"Vykresleni Bezierovy plochy zadane ridici siti. Parametry stejne jako u BezierSurface. Pouziti: PlotBezierSurface[{ridici sit},{parametr1,parametr2}]";\


RatBezierSurface::usage = \
"Vypocte parametrizaci racinalni Bezierovy plochy pro zadanou ridici sit. Ridici sit se zadava jako trojrozmerne pole typu {{{_,_,_,_}..}..}, ktere predstavuje matici bodu ridici site {{P00,...,P0n},...,{Pm0,...,Pmn}} a kazde Pij je vektor o ctyrech souradnicich - prvni tri predstavuji kartezske souradnice bodu, ctvrta predstavuje vahu daneho bodu. Druhym parametrem je vektor o dvou slozkach, obsahujici symboly pouzite pro parametry vysledne parametrizace. Pouziti: RatBezierSurface[{ridici sit},{parametr1,parametr2}]";\


PlotRatBezierSurface::usage = \
"Vykresleni racionalni Bezierovy plochy zadane ridici siti. Parametry stejne jako u RatBezierSurface. Pouziti: PlotRatBezierSurface[{ridici sit},{{parametr1,parametr2}]";\


NBasis::usage = \
"Vypocet B-Spline baze pro zadany vektor parametrizace. Druhym parametrem funkce je symbol parametru vysledne baze. Pouziti: NBasis[vektor parametrizace,parametr]";\


PlotNBasis::usage =
    "Vykresleni B-Spline bazovych funkci pro zadany vektor parametrizace. Pouziti: PlotNBasis[vektor parametrizace,parametr]";\


BSplineCurve::usage = 
    "Vypocte parametrizaci rovinne (body ridici polygonu maji dve souradnice) i prostorove (body ridici polygonu maji tri souradnice) B-Spline krivky pro zadany ridici polygon a vektor parametrizace. Symbol parametru se zadava jako posledni parametr funkce. Pouziti: BSplineCurve[{body ridiciho polygonu},vektor parametrizace,promenna parametrizace]";\


PlotBSplineCurve::usage = \
"Vykresleni rovinne i prostorove B-Spline krivky zadane ridicim polygonem a vektorem parametrizace. Parametry stejne jako u BSplineCurve, navic je pouze posledni parametr urcujici barvu krivku (cislo v rozsahu 0, 1). Pouziti: PlotBSplineCurve[{body ridiciho polygonu},vektor parametrizace,promenna parametrizace,barva]";\


BSplineSurface::usage = \
"Vypocte parametrizaci B-Spline plochy pro zadanou ridici sit a dva vektory parametrizace. Ridici sit se zadava jako trojrozmerne pole typu {{{_,_,_}..}..}, ktere predstavuje matici bodu ridici site {{P00,...,P0n},...,{Pm0,...,Pmn}} a kazde Pij je bod v prostoru. Druhym parametrem jsou vektory parametrizace a poslednim parametrem je vektor o dvou slozkach, obsahujici symboly pouzite pro parametry vysledne parametrizace. Pouziti: BSplineSurface[{ridici sit},{vektor parametrizace 1,vektor parametrizace 2},{parametr1,parametr2}]";\


PlotBSplineSurface::usage = \
"Vykresleni B-Spline plochy zadane ridici siti a vektory parametrizace. Parametry stejne jako u BSplineSurface. Pouziti: PlotBSplineSurface[{ridici sit},{vektor parametrizace 1,vektor parametrizace 2},{parametr1,parametr2}]";\


NURBSCurve::usage = \
"Vypocte parametrizaci rovinne (body ridici polygonu maji tri souradnice, posledni slozka je vaha bodu) i prostorove (body ridici polygonu maji ctyri souradnice, posledni souradnice ja vaha) NURBS krivky pro zadany ridici polygon a vektor parametrizace. Symbol parametru se zadava jako posledni parametr funkce. Pouziti: NURBSCurve[{body ridiciho polygonu},vektor parametrizace,promenna parametrizace]";\


PlotNURBSCurve::usage = 
    "Vykresleni rovinne i prostorove NURBS krivky zadane ridicim polygonem a vektorem parametrizace. Parametry stejne jako u NURBSCurve, navic je pouze posledni parametr urcujici barvu krivku (cislo v rozsahu 0, 1). Pouziti: PlotNURBSCurve[{body ridiciho polygonu},vektor parametrizace,promenna parametrizace,barva]";\


PlotNURBSCurveClassic::usage = \
"Podobne jako PlotNURBSCurve, pouze se jinak zobrazuje ridici polygon v pripade, ze obsahuje ridici body s nulovou vahou.";\


NURBSSurface::usage = \
"Vypocte parametrizaci NURBS plochy pro zadanou ridici sit a dva vektory parametrizace. Ridici sit se zadava jako trojrozmerne pole typu {{{_,_,_}..}..}, ktere predstavuje matici bodu ridici site {{P00,...,P0n},...,{Pm0,...,Pmn}} a kazde Pij je vektor o ctyrech souradnicich - prvni tri predstavuji kartezske souradnice bodu, ctvrta predstavuje vahu daneho bodu. Druhym parametrem jsou vektory parametrizace a poslednim parametrem je vektor o dvou slozkach, obsahujici symboly pouzite pro parametry vysledne parametrizace. Pouziti: NURBSSurface[{ridici sit},{vektor parametrizace 1,vektor parametrizace 2},{parametr1,parametr2}]";\


PlotNURBSSurface::usage = \
"Vykresleni NURBS plochy zadane ridici siti a vektory parametrizace. Parametry stejne jako u NURBSSurface. Pouziti: PlotNURBSSurface[{ridici sit},{vektor parametrizace 1,vektor parametrizace 2},{parametr1,parametr2}]";\


Rotation2D::usage = \
"Rotace mnoziny bodu v rovine o zadany uhel alfa. Body jsou zadavany jako matice, kde radek matice predstavuje souradnice jednoho bodu, a to bud kartezske souradnice nebo homogenni souradnice. Pouziti: Rotation2D[{body},uhel]";\


Reflection2D::usage = \
"Symetrie v rovine podle zadane primky. Body jsou zadavany jako matice, kde radek matice predstavuje souradnice jednoho bodu, a to bud kartezske souradnice nebo homogenni souradnice. Osa symetrie je zadavana pomoci koefiecientu {a,b,c} obecne rovnice primky v rovine ax+by+c=0. Pouziti: Reflection2D[{body},{a,b,c}]";\


Reflection3D::usage = 
    "Symetrie v prostoru podle zadane roviny. Body jsou zadavany jako matice, kde radek matice predstavuje souradnice jednoho bodu, a to bud kartezske souradnice nebo homogenni souradnice. Rovina symetrie je zadavana pomoci koefiecientu {a,b,c,d} obecne rovnice roviny v prostoru ax+by+cz+d=0. Pouziti: Reflection3D[{body},{a,b,c,d}]";\


Rotation3D::usage = \
"Rotace mnoziny bodu v prostoru kolem zvolene primky o zadany uhel alfa. Body jsou zadavany jako matice, kde radek matice predstavuje souradnice jednoho bodu, a to bud kartezske souradnice nebo homogenni souradnice. Primka je zadavana pomoci smeroveho vektoru a jednoho bodu. Pouziti: Rotation2D[{body},{smerovy vektor,bod},uhel]";\


RevolutionSurface::usage = \
"Vypocte parametrizaci rotacni plochy, vznikle rotaci zadane NURBS krivky okolo zvolene primky. Vstupem jsou body ridiciho polygonu NURBS krivky (zadane stejnym zpusobem jako pro NURBSCurve), vektor parametrizace teto krivky, vektor obsahujici parametry vysledne parametrizace a poslednim parametrem je primka, okolo ktere ma zvolena krivka rotovat (zadana pomoci smeroveho vektoru a jednoho jejiho bodu). Pouziti: RevolutionSurface[{body krivky},vektor parametrizace,{parametr1,parametr2},{smerovy vektor,bod primky}]";\


PlotRevolutionSurface::usage = \
"Vykresleni rotacni plochy, vznikle rotaci zadane NURBS krivky okolo zvolene primky. Parametry stejne jako u RevolutionSurface. Pouziti: PlotRevolutionSurface[{body krivky},vektor parametrizace,{parametr1,parametr2},{smerovy vektor,bod primky}]";\


Begin["CAGD`Private`"]

NormalizeVectors[A : {{_, _} ..}] := Map[#/Norm[#] &, A];

ChordalParametrization[Pts_] := Module[{},
      NotNormalized = 
        FoldList[Plus, 0, 
          Table[Norm[Pts[[i + 1]] - Pts[[i]]], {i, 1, Length[Pts] - 1}]];
      Return[NotNormalized/Last[NotNormalized]]];

Ferguson[P0_, P1_, P0d_, P1d_, Parameters_, t_] := Module[{},
      kp = Apply[Subtract, Reverse[Parameters]];
      H = {2/kp^3 s^3 - 3/kp^2 s^2 + 1, -2/kp^3 s^3 + 3/kp^2 s^2, 
          s^3/kp^2 - 2/kp s^2 + s, s^3/kp^2 - s^2/kp};
      P = Simplify[(H /. {s -> (t - Parameters[[1]])}).{P0, P1, P0d, P1d}];
      Return[P]];

Options[CubicSpline] = {Parametrization -> Classic, 
      BoundaryConditions -> Natural, Color -> 0.6, 
      DisplayFunction -> $DisplayFunction};

CubicSpline[Pts : {{_, _} ..}, Prom_, opts___] := Module[{},
      n = Length[Pts] - 1;
      OptionRules = Select[{opts}, MatchQ[#, _Rule] &];
      BCond = Flatten[Complement[{opts}, OptionRules], 1];
      T = 
        If[(Parametrization /. OptionRules /. Options[CubicSpline][[1]]) === 
            Chordal, ChordalParametrization[Pts], Range[0, n]/n];
      K = Drop[T - RotateRight[T], 1];
      SystemMatrix = 
        Map[PadRight[
              PadLeft[{1/K[[# + 1]], 2/K[[# + 1]] + 2/K[[# + 2]], 
                  1/K[[# + 2]]}, 3 + #], n + 1] &, Range[0, n - 2]];
      SystemRightSide = 
        Map[3/K[[# + 2]]^2 Pts[[# + 3]] + (3/K[[# + 1]]^2 - 
                    3/K[[# + 2]]^2) Pts[[# + 2]] - 
              3/K[[# + 1]]^2 Pts[[# + 1]] &, Range[0, n - 2]];
      Switch[BoundaryConditions /. OptionRules /. Options[CubicSpline][[2]],
        Clamped, Block[{},
          Print["Clamped boundary conditions ..."];
          
          SystemMatrix = 
            Transpose[Delete[Transpose[SystemMatrix], {{1}, {-1}}]];
          SystemRightSide[[1]] = SystemRightSide[[1]] - BCond[[1]]/K[[1]];
          
          SystemRightSide[[n - 1]] = 
            SystemRightSide[[n - 1]] - BCond[[2]]/K[[n]];
          
          TangentVectors = 
            Append[Prepend[LinearSolve[SystemMatrix, SystemRightSide], 
                BCond[[1]]], BCond[[2]]];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, Prom], 
                  Join[{Prom}, {T[[#]], T[[# + 1]]}]] &, Range[1, n]];
          ],
        Periodic, Block[{},
          Print["Closed cubic spline ..."];
          
          T = If[(Parametrization /. OptionRules /. 
                    Options[CubicSpline][[1]]) === Chordal, 
              ChordalParametrization[Join[Pts, {Pts[[1]]}]], 
              Range[0, n + 1]/(n + 1)];
          K = Drop[T - RotateRight[T], 1];
          SystemMatrix = 
            Map[PadRight[
                  PadLeft[{1/K[[# + 1]], 2/K[[# + 1]] + 2/K[[# + 2]], 
                      1/K[[# + 2]]}, 3 + #], n + 1] &, Range[0, n - 2]];
          
          SystemRightSide = 
            Map[3/K[[# + 2]]^2 Pts[[# + 3]] + (3/K[[# + 1]]^2 - 
                        3/K[[# + 2]]^2) Pts[[# + 2]] - 
                  3/K[[# + 1]]^2 Pts[[# + 1]] &, Range[0, n - 2]];
          
          LastRow = 
            Prepend[PadLeft[{1/K[[n]], 2/K[[n]] + 2/K[[n + 1]]}, n], 
              1/K[[n + 1]]];
          
          FirstRow = 
            Append[PadRight[{2/K[[n + 1]] + 2/K[[1]], 1/K[[1]]}, n], 
              1/K[[n + 1]]];
          SystemMatrix = Append[Prepend[SystemMatrix, FirstRow], LastRow];
          
          LastElement = 
            3/K[[n + 1]]^2 Pts[[1]] + (3/K[[n]]^2 - 
                    3/K[[n + 1]]^2) Pts[[n + 1]] - 3/K[[n]]^2 Pts[[n]];
          
          FirstElement = 
            3/K[[1]]^2 Pts[[2]] + (3/K[[n + 1]]^2 - 3/K[[1]]^2) Pts[[1]] - 
              3/K[[n + 1]]^2 Pts[[n + 1]];
          
          SystemRightSide = 
            Append[Prepend[SystemRightSide, FirstElement], LastElement];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          PtsNew = Join[Pts, {Pts[[1]]}];
          TangentVectorsNew = Join[TangentVectors, {TangentVectors[[1]]}];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[PtsNew[[#]], PtsNew[[# + 1]], 
                    TangentVectorsNew[[#]], 
                    TangentVectorsNew[[# + 1]], {T[[#]], T[[# + 1]]}, Prom], 
                  Join[{Prom}, {T[[#]], T[[# + 1]]}]] &, Range[1, n + 1]];
          ],
        Natural, Block[{},
          Print["Natural cubic spline ..."];
          
          SystemMatrix = 
            Join[{PadRight[{2, 1}, n + 1]}, 
              SystemMatrix, {PadLeft[{1, 2}, n + 1]}];
          
          SystemRightSide = 
            Join[{3/K[[1]] (Pts[[2]] - Pts[[1]])}, 
              SystemRightSide, {3/K[[n]] (Pts[[n + 1]] - Pts[[n]])}];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, Prom], 
                  Join[{Prom}, {T[[#]], T[[# + 1]]}]] &, Range[1, n]];
          ],
        SecondDerivatives, Block[{},
          Print["Given second derivatives in the first and last points ..."];
          
          SystemMatrix = 
            Join[{PadRight[{2, 1}, n + 1]}, 
              SystemMatrix, {PadLeft[{1, 2}, n + 1]}];
          
          SystemRightSide = 
            Join[{3/K[[1]] (Pts[[2]] - Pts[[1]]) - K[[1]]/2 BCond[[1]]}, 
              SystemRightSide, {3/K[[n]] (Pts[[n + 1]] - Pts[[n]]) - 
                  K[[n]]/2 BCond[[2]]}];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, Prom], 
                  Join[{Prom}, {T[[#]], T[[# + 1]]}]] &, Range[1, n]];
          ]];
      Return[Simplify[ArcParametrizations]]];

PlotCubicSpline[Pts : {{_, _} ..}, Prom_, opts___] := Module[{},
      n = Length[Pts] - 1;
      OptionRules = Select[{opts}, MatchQ[#, _Rule] &];
      BCond = Flatten[Complement[{opts}, OptionRules], 1];
      T = 
        If[(Parametrization /. OptionRules /. Options[CubicSpline][[1]]) === 
            Chordal, ChordalParametrization[Pts], Range[0, n]/n];
      K = Drop[T - RotateRight[T], 1];
      SystemMatrix = 
        Map[PadRight[
              PadLeft[{1/K[[# + 1]], 2/K[[# + 1]] + 2/K[[# + 2]], 
                  1/K[[# + 2]]}, 3 + #], n + 1] &, Range[0, n - 2]];
      SystemRightSide = 
        Map[3/K[[# + 2]]^2 Pts[[# + 3]] + (3/K[[# + 1]]^2 - 
                    3/K[[# + 2]]^2) Pts[[# + 2]] - 
              3/K[[# + 1]]^2 Pts[[# + 1]] &, Range[0, n - 2]];
      Switch[BoundaryConditions /. OptionRules /. Options[CubicSpline][[2]],
        Clamped, Block[{},
          Print["Clamped boundary conditions ..."];
          
          SystemMatrix = 
            Transpose[Delete[Transpose[SystemMatrix], {{1}, {-1}}]];
          SystemRightSide[[1]] = SystemRightSide[[1]] - BCond[[1]]/K[[1]];
          
          SystemRightSide[[n - 1]] = 
            SystemRightSide[[n - 1]] - BCond[[2]]/K[[n]];
          
          TangentVectors = 
            Append[Prepend[LinearSolve[SystemMatrix, SystemRightSide], 
                BCond[[1]]], BCond[[2]]];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, 
                    Prom], {T[[#]], T[[# + 1]]}] &, Range[1, n]];
          
          Krivky = 
            Map[ParametricPlot[Drop[ArcParametrizations[[#]], -1], 
                  Evaluate[Join[{Prom}, T[[{#, # + 1}]]]], 
                  PlotStyle -> 
                    Hue[Color /. OptionRules /. Options[CubicSpline]], 
                  DisplayFunction -> Identity] &, Range[1, n]];
          ],
        Periodic, Block[{},
          Print["Closed cubic spline ..."];
          
          T = If[(Parametrization /. OptionRules /. 
                    Options[CubicSpline][[1]]) === Chordal, 
              ChordalParametrization[Join[Pts, {Pts[[1]]}]], 
              Range[0, n + 1]/(n + 1)];
          K = Drop[T - RotateRight[T], 1];
          
          SystemMatrix = 
            Map[PadRight[
                  PadLeft[{1/K[[# + 1]], 2/K[[# + 1]] + 2/K[[# + 2]], 
                      1/K[[# + 2]]}, 3 + #], n + 1] &, Range[0, n - 2]];
          
          SystemRightSide = 
            Map[3/K[[# + 2]]^2 Pts[[# + 3]] + (3/K[[# + 1]]^2 - 
                        3/K[[# + 2]]^2) Pts[[# + 2]] - 
                  3/K[[# + 1]]^2 Pts[[# + 1]] &, Range[0, n - 2]];
          
          LastRow = 
            Prepend[PadLeft[{1/K[[n]], 2/K[[n]] + 2/K[[n + 1]]}, n], 
              1/K[[n + 1]]];
          
          FirstRow = 
            Append[PadRight[{2/K[[n + 1]] + 2/K[[1]], 1/K[[1]]}, n], 
              1/K[[n + 1]]];
          SystemMatrix = Append[Prepend[SystemMatrix, FirstRow], LastRow];
          
          LastElement = 
            3/K[[n + 1]]^2 Pts[[1]] + (3/K[[n]]^2 - 
                    3/K[[n + 1]]^2) Pts[[n + 1]] - 3/K[[n]]^2 Pts[[n]];
          
          FirstElement = 
            3/K[[1]]^2 Pts[[2]] + (3/K[[n + 1]]^2 - 3/K[[1]]^2) Pts[[1]] - 
              3/K[[n + 1]]^2 Pts[[n + 1]];
          
          SystemRightSide = 
            Append[Prepend[SystemRightSide, FirstElement], LastElement];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          PtsNew = Join[Pts, {Pts[[1]]}];
          TangentVectorsNew = Join[TangentVectors, {TangentVectors[[1]]}];
          
          ArcParametrizations = 
            Map[
              Append[Ferguson[PtsNew[[#]], PtsNew[[# + 1]], 
                    TangentVectorsNew[[#]], 
                    TangentVectorsNew[[# + 1]], {T[[#]], T[[# + 1]]}, 
                    Prom], {T[[#]], T[[# + 1]]}] &, Range[1, n + 1]];
          
          Krivky = 
            Map[ParametricPlot[Drop[ArcParametrizations[[#]], -1], 
                  Evaluate[Join[{Prom}, T[[{#, # + 1}]]]], 
                  PlotStyle -> 
                    Hue[Color /. OptionRules /. Options[CubicSpline]], 
                  DisplayFunction -> Identity] &, Range[1, n + 1]];
          ],
        Natural, Block[{},
          Print["Natural cubic spline ..."];
          
          SystemMatrix = 
            Join[{PadRight[{2, 1}, n + 1]}, 
              SystemMatrix, {PadLeft[{1, 2}, n + 1]}];
          
          SystemRightSide = 
            Join[{3/K[[1]] (Pts[[2]] - Pts[[1]])}, 
              SystemRightSide, {3/K[[n]] (Pts[[n + 1]] - Pts[[n]])}];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, 
                    Prom], {T[[#]], T[[# + 1]]}] &, Range[1, n]];
          
          Krivky = 
            Map[ParametricPlot[Drop[ArcParametrizations[[#]], -1], 
                  Evaluate[Join[{Prom}, T[[{#, # + 1}]]]], 
                  PlotStyle -> 
                    Hue[Color /. OptionRules /. Options[CubicSpline]], 
                  DisplayFunction -> Identity] &, Range[1, n]];
          ],
        SecondDerivatives, Block[{},
          Print["Given second derivatives in the first and last points ..."];
          
          SystemMatrix = 
            Join[{PadRight[{2, 1}, n + 1]}, 
              SystemMatrix, {PadLeft[{1, 2}, n + 1]}];
          
          SystemRightSide = 
            Join[{3/K[[1]] (Pts[[2]] - Pts[[1]]) - K[[1]]/2 BCond[[1]]}, 
              SystemRightSide, {3/K[[n]] (Pts[[n + 1]] - Pts[[n]]) - 
                  K[[n]]/2 BCond[[2]]}];
          TangentVectors = LinearSolve[SystemMatrix, SystemRightSide];
          
          ArcParametrizations = 
            Map[Append[
                  Ferguson[Pts[[#]], Pts[[# + 1]], TangentVectors[[#]], 
                    TangentVectors[[# + 1]], {T[[#]], T[[# + 1]]}, 
                    Prom], {T[[#]], T[[# + 1]]}] &, Range[1, n]];
          Krivky = 
            Map[ParametricPlot[Drop[ArcParametrizations[[#]], -1], 
                  Evaluate[Join[{Prom}, T[[{#, # + 1}]]]], 
                  PlotStyle -> 
                    Hue[Color /. OptionRules /. Options[CubicSpline]], 
                  DisplayFunction -> Identity] &, Range[1, n]];
          ]];
      Body = 
        ListPlot[Pts, PlotStyle -> PointSize[0.03], PlotJoined -> False, 
          DisplayFunction -> Identity];
      Vektory = 
        ListPlotVectorField[
          Transpose[{Pts, NormalizeVectors[TangentVectors]}], 
          DisplayFunction -> Identity];
      Obrazek = 
        Show[Krivky, Body, Vektory, 
          DisplayFunction -> (DisplayFunction /. OptionRules /. 
                Options[CubicSpline])];
      Return[Obrazek]];

Options[QuadraticSpline] = {Parametrization -> Classic, Color -> 0.6, 
      DisplayFunction -> $DisplayFunction};

QuadraticSpline[Pts : {{_, _} ..}, BCond : {_, _}, Prom_, opts___] := 
    Module[{n = Length[Pts] - 1},
      T = 
        If[(Parametrization /. {opts} /. Options[QuadraticSpline]) === 
            Chordal, ChordalParametrization[Pts], Range[0, n]/n];
      K = Drop[T - RotateRight[T], 1];
      G = {-1/k^2 s^2 + 1, 1/k^2 s^2, -1/k s^2 + s};
      ParsCurrent = FoldList[Module[{},
              
              F = Simplify[(G /. {s -> (Prom - T[[#2]]), 
                          k -> K[[#2]]}).{Pts[[#2]], 
                      Pts[[#2 + 1]], #1[[2]]}];
              TangentNew = D[F, Prom] /. {Prom -> T[[#2 + 1]]};
              {F, TangentNew}] &, {0, BCond}, Range[n]];
      ArcParametrizations = 
        Map[Append[Drop[ParsCurrent[[All, 1]], 1][[#]], 
              Join[{Prom}, {T[[#]], T[[# + 1]]}]] &, Range[n]];
      TangentVectors = ParsCurrent[[All, 2]];
      Return[ArcParametrizations]];

PlotQuadraticSpline[Pts : {{_, _} ..}, BCond : {_, _}, Prom_, opts___] := 
    Module[{n = Length[Pts] - 1},
      T = 
        If[(Parametrization /. {opts} /. Options[QuadraticSpline]) === 
            Chordal, ChordalParametrization[Pts], Range[0, n]/n];
      K = Drop[T - RotateRight[T], 1];
      G = {-1/k^2 s^2 + 1, 1/k^2 s^2, -1/k s^2 + s};
      ParsCurrent = FoldList[Module[{},
              
              F = Simplify[(G /. {s -> (Prom - T[[#2]]), 
                          k -> K[[#2]]}).{Pts[[#2]], 
                      Pts[[#2 + 1]], #1[[2]]}];
              TangentNew = D[F, Prom] /. {Prom -> T[[#2 + 1]]};
              {F, TangentNew}] &, {0, BCond}, Range[n]];
      ArcParametrizations = 
        Map[Append[
              Drop[ParsCurrent[[All, 1]], 1][[#]], {T[[#]], T[[# + 1]]}] &, 
          Range[n]];
      TangentVectors = ParsCurrent[[All, 2]];
      Krivky = 
        Map[ParametricPlot[Drop[ArcParametrizations[[#]], -1], 
              Evaluate[Join[{Prom}, T[[{#, # + 1}]]]], 
              PlotStyle -> Hue[Color /. {opts} /. Options[QuadraticSpline]], 
              DisplayFunction -> Identity] &, Range[1, n]];
      Body = 
        ListPlot[Pts, PlotStyle -> PointSize[0.03], PlotJoined -> False, 
          DisplayFunction -> Identity];
      Vektory = 
        ListPlotVectorField[
          Transpose[{Pts, NormalizeVectors[TangentVectors]}], 
          DisplayFunction -> Identity];
      Obrazek = 
        Show[Krivky, Body, Vektory, 
          DisplayFunction -> (DisplayFunction /. {opts} /. 
                Options[QuadraticSpline])];
      Return[Obrazek]];

Bernstein[i_, n_, v_] := Binomial[n, i] (1 - v)^(n - i) v^i;

BezierCurve[pt : {{_, _} ..} | {{_, _, _} ..}, v_] := 
    Module[{n = Length[pt] - 1},
      Simplify[Table[Bernstein[i, n, v], {i, 0, n}].pt]];

Options[PlotBezierCurve] = {DisplayFunction -> $DisplayFunction, 
      Color -> 0.6};

PlotBezierCurve[pt : {{_, _} ..} | {{_, _, _} ..}, v_, Opt___] := 
    Module[{BC},
      BC = BezierCurve[pt, v];
      If[Length[pt[[1]]] == 2, Module[{},
          
          CurveGraph = 
            ParametricPlot[BC, {v, 0, 1}, PlotRange -> All, 
              PlotStyle -> Hue[Color /. {Opt} /. Options[PlotBezierCurve]], 
              DisplayFunction -> Identity];
          
          CurvePolygon = 
            ListPlot[pt, PlotJoined -> True, DisplayFunction -> Identity];
          
          CurvePoints = 
            ListPlot[pt, PlotJoined -> False, PlotStyle -> PointSize[0.02], 
              DisplayFunction -> Identity]], Module[{},
          CurveGraph = 
            ParametricPlot3D[
              Join[BC, {Hue[Color /. {Opt} /. Options[PlotBezierCurve]]}], {v,
                 0, 1}, DisplayFunction -> Identity];
          CurvePolygon = Graphics3D[Line[pt]];
          CurvePoints = Graphics3D[Map[{PointSize[0.02], Point[#]} &, pt]]
          ]];
      Show[CurveGraph, CurvePolygon, CurvePoints, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotBezierCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

RatBezierCurve[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, v_] := 
    Module[{n = Length[pt] - 1},
      If[Length[First[pt]] == 3, 
        ptw = pt /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ} :> 
                If[c == 0, {a, b, c}, {a c, b c, c}]}, 
        ptw = pt /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ, d_?NumberQ} :> 
                If[d == 0, {a, b, c, d}, {a d, b d, c d, d}]}];
      ParPom = Simplify[Table[Bernstein[i, n, v], {i, 0, n}].ptw];
      Return[Drop[ParPom/Last[ParPom], -1]]];

Options[PlotRatBezierCurve] = {DisplayFunction -> $DisplayFunction, 
      Color -> 0.6};

PlotRatBezierCurveClassic[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, v_, 
      Opt___] := Module[{RBC},
      RBC = RatBezierCurve[pt, v];
      If[Length[pt[[1]]] == 3, Module[{},
          
          CurveGraph = 
            ParametricPlot[RBC, {v, 0, 1}, 
              PlotStyle -> Hue[Color /. {Opt} /. Options[PlotRatBezierCurve]],
               DisplayFunction -> Identity];
          
          CurvePolygon = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> True, 
              DisplayFunction -> Identity];
          
          CurvePoints = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> False, 
              PlotStyle -> PointSize[0.02], DisplayFunction -> Identity];
          
          CurveWeights = 
            Graphics[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]],
        Module[{},
          
          CurveGraph = 
            ParametricPlot3D[Join[RBC, {Hue[Color]}], {v, 0, 1}, 
              PlotStyle -> Hue[Color /. {Opt} /. Options[PlotRatBezierCurve]],
               DisplayFunction -> Identity];
          CurvePolygon = Graphics3D[Line[Map[Drop[#, -1] &, pt]]];
          
          CurvePoints = 
            Graphics3D[
              Map[{PointSize[0.02], Point[#]} &, Map[Drop[#, -1] &, pt]]]; 
          CurveWeights = 
            Graphics3D[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1]]} &, pt]]]];
      Show[CurveGraph, CurvePolygon, CurvePoints, CurveWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotRatBezierCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

PlotRatBezierCurve[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, v_, Opt___] := 
    Module[{RBC},
      RBC = RatBezierCurve[pt, v];
      If[Length[pt[[1]]] == 3, Module[{},
          
          CurveGraph = 
              ParametricPlot[RBC, {v, 0, 1}, 
                PlotStyle -> 
                  Hue[Color /. {Opt} /. Options[PlotRatBezierCurve]], 
                DisplayFunction -> Identity];;
          PN = Flatten[Position[pt, _?(#[[3]] == 0 &), 1, Heads -> False]];
          PolygonPoints = If[Length[PN] > 0,
              
              Append[Prepend[
                  Table[Append[
                      Prepend[pt[[Range[PN[[i]] + 1, PN[[i + 1]] - 1]]], 
                        pt[[PN[[i]]]] + pt[[PN[[i]] + 1]]], 
                      pt[[PN[[i + 1]] - 1]] + pt[[PN[[i + 1]]]]], {i, 1, 
                      Length[PN] - 1}], 
                  Append[pt[[Range[PN[[1]] - 1]]], 
                    pt[[PN[[1]] - 1]] + pt[[PN[[1]]]]]], 
                Prepend[pt[[Range[Last[PN] + 1, Length[pt]]]], 
                  pt[[Last[PN]]] + pt[[Last[PN] + 1]]]], {pt}];
          
          CurvePolygon = 
            Show[Map[
                ListPlot[Transpose[Drop[Transpose[#], -1]], 
                    PlotJoined -> True, DisplayFunction -> Identity] &, 
                PolygonPoints]];
          
          CurvePoints = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> False, 
              PlotStyle -> PointSize[0.02], DisplayFunction -> Identity];
          
          CurveWeights = 
            Graphics[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]], Module[{},
          
          CurveGraph = 
              ParametricPlot3D[
                Join[RBC, {Hue[
                      Color /. {Opt} /. Options[PlotRatBezierCurve]]}], {v, 0,
                   1}, PlotStyle -> Hue[1], DisplayFunction -> Identity];;
          PN = Flatten[Position[pt, _?(#[[4]] == 0 &), 1, Heads -> False]];
          
          PolygonPoints = 
            If[Length[PN] > 0, 
              Append[Prepend[
                  Table[Append[
                      Prepend[pt[[Range[PN[[i]] + 1, PN[[i + 1]] - 1]]], 
                        pt[[PN[[i]]]] + pt[[PN[[i]] + 1]]], 
                      pt[[PN[[i + 1]] - 1]] + pt[[PN[[i + 1]]]]], {i, 1, 
                      Length[PN] - 1}], 
                  Append[pt[[Range[PN[[1]] - 1]]], 
                    pt[[PN[[1]] - 1]] + pt[[PN[[1]]]]]], 
                Prepend[pt[[Range[Last[PN] + 1, Length[pt]]]], 
                  pt[[Last[PN]]] + pt[[Last[PN] + 1]]]], {pt}];
          
          CurvePolygon = 
            Show[Map[
                Graphics3D[
                    Line[Transpose[Drop[Transpose[#], -1]], 
                      PlotJoined -> True, DisplayFunction -> Identity]] &, 
                PolygonPoints]];
          
          CurvePoints = 
            Graphics3D[
              Map[{PointSize[0.02], Point[#]} &, Map[Drop[#, -1] &, pt]]];
          
          CurveWeights = 
            Graphics3D[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]
          ]];
      Show[CurveGraph, CurvePolygon, CurvePoints, CurveWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotRatBezierCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

BezierSurface[pts : {{{_, _, _} ..} ..}, {v1_, v2_}] := 
    Module[{m = Length[pts] - 1, n = Length[First[pts]] - 1, U, V},
      {U, V} = 
        MapThread[
          Table[Bernstein[i, #1, #2], {i, 0, #1}] &, {{m, n}, {v1, v2}}];
      Simplify[Apply[Plus, (U.pts) V]]];

Options[PlotBezierSurface] = {DisplayFunction -> $DisplayFunction};

PlotBezierSurface[pts : {{{_, _, _} ..} ..}, {v1_, v2_}, Opt___] := 
    Module[{},
      BS = BezierSurface[pts, {v1, v2}];
      SurfaceGraph = 
        ParametricPlot3D[BS, {v1, 0, 1}, {v2, 0, 1}, 
          DisplayFunction -> Identity];
      SurfaceNet = Graphics3D[Map[Line[#] &, Join[pts, Transpose[pts]]]];
      SurfacePoints = 
        Graphics3D[Map[{PointSize[0.015], Point[#]} &, Flatten[pts, 1]]];
      Show[SurfaceGraph, SurfaceNet, SurfacePoints, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotBezierSurface]), Opt]];

RatBezierSurface[pts : {{{_, _, _, _} ..} ..}, {v1_, v2_}] := 
    Module[{m = Length[pts] - 1, n = Length[First[pts]] - 1, U, V},
      {U, V} = 
        MapThread[
          Table[Bernstein[i, #1, #2], {i, 0, #1}] &, {{m, n}, {v1, v2}}];
      ptsw = 
        pts /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ, d_?NumberQ} :> 
              If[d == 0, {a, b, c, d}, {a d, b d, c d, d}]};
      ParPom = Simplify[Apply[Plus, (U.ptsw) V]];
      Return[Drop[ParPom/Last[ParPom], -1]]];

Options[PlotRatBezierSurface] = {DisplayFunction -> $DisplayFunction};

PlotRatBezierSurface[pts : {{{_, _, _, _} ..} ..}, {v1_, v2_}, Opt___] := 
    Module[{},
      RBS = RatBezierSurface[pts, {v1, v2}];
      SurfaceGraph = 
        ParametricPlot3D[RBS, {v1, 0, 1}, {v2, 0, 1}, 
          DisplayFunction -> Identity];
      SurfaceNet = 
        Graphics3D[
          Map[Line[Map[Drop[#, -1] &, #]] &, Join[pts, Transpose[pts]]]];
      SurfacePoints = 
        Graphics3D[
          Map[{PointSize[0.015], Point[Drop[#, -1]]} &, Flatten[pts, 1]]];
      SurfaceWeights = 
        Graphics3D[
          Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1]]} &, 
            Flatten[pts, 1]]];
      Show[SurfaceGraph, SurfaceNet, SurfacePoints, SurfaceWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotRatBezierSurface]), Opt]];

NBasis[U_List, u_] := Module[{},
      DegU = Length[DeleteCases[U, a_?Positive]] - 1;
      NBStep[X_, W_, p_] := 
        Map[If[X[[#1 + p]] - X[[#1]] == 0, 
                0, (u - X[[#1]])/(X[[#1 + p]] - X[[#1]]) W[[#1]]] + 
              If[X[[#1 + p + 1]] - X[[#1 + 1]] == 0, 
                0, (X[[#1 + p + 1]] - u)/(X[[#1 + p + 1]] - 
                        X[[#1 + 1]]) W[[#1 + 1]]] &, 
          Table[j, {j, 1, Length[W] - 1}]];
      NBasisPom[XX_, WW_, pp_] := 
        Map[NBStep[XX, WW[[#]], pp] &, Table[i, {i, 1, Length[WW]}]];
      P = 
        Fold[NBasisPom[U, #1, #2] &, 
          Map[ReplacePart[Table[0, {i, Length[U] - 1}], 1, #] &, 
            Flatten[Position[Rest[U] - Take[U, Length[U] - 1], a_?Positive]]],
           Table[i, {i, 1, DegU}]];
      Return[P]];

Options[PlotNBasis] = {DisplayFunction -> $DisplayFunction};

PlotNBasis[U_List, u_, Opt___] := Module[{},
      Nbaze = NBasis[U, u];
      BasisGraph = 
        Table[Plot[Evaluate[Nbaze[[i]]], 
            Evaluate[Join[{u}, Union[U][[{i, i + 1}]]]], 
            DisplayFunction -> Identity], {i, 1, Length[Nbaze]}];
      Show[BasisGraph, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotRatBezierSurface]), Opt]];

BSplineCurve[pt : {{_, _} ..} | {{_, _, _} ..}, U_, v_] := Module[{},
      CurveBasis = NBasis[U, v];
      P = 
        Map[Append[Simplify[CurveBasis.pt][[#]], 
              Join[{v}, Part[Union[U], {#, # + 1}]]] &, 
          Table[i, {i, 1, Length[Union[U]] - 1}]];
      Return[P]];

Options[PlotBSplineCurve] = {DisplayFunction -> $DisplayFunction, 
      Color -> 0.6};

PlotBSplineCurve[pt : {{_, _} ..} | {{_, _, _} ..}, U_, v_, Opt___] := 
    Module[{BSC},
      BSC = BSplineCurve[pt, U, v];
      If[Length[pt[[1]]] == 2, Module[{},
          
          CurveGraph = 
            Map[ParametricPlot[Drop[#, -1], Evaluate[Last[#]], 
                  PlotStyle -> 
                    Hue[Color /. {Opt} /. Options[PlotRatBezierCurve]], 
                  DisplayFunction -> Identity] &, BSC];
          
          CurvePolygon = 
            ListPlot[pt, PlotJoined -> True, DisplayFunction -> Identity];
          
          CurvePoints = 
            ListPlot[pt, PlotJoined -> False, PlotStyle -> PointSize[0.02], 
              DisplayFunction -> Identity]], Module[{},
          
          CurveGraph = 
            Map[ParametricPlot3D[
                  Join[Drop[#, -1], {Hue[
                        Color /. {Opt} /. Options[PlotRatBezierCurve]]}], 
                  Evaluate[Last[#]], DisplayFunction -> Identity] &, BSC];
          CurvePolygon = Graphics3D[Line[pt]];
          CurvePoints = Graphics3D[Map[{PointSize[0.02], Point[#]} &, pt]]
          ]];
      Show[CurveGraph, CurvePolygon, CurvePoints, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotBSplineCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

BSplineSurface[pts : {{{_, _, _} ..} ..}, {U1_, U2_}, {v1_, v2_}] := 
    Module[{},
      SurfaceBasis1 = NBasis[U1, v1];
      SurfaceBasis2 = NBasis[U2, v2];
      DegU1 = Length[DeleteCases[U1, a_?Positive]] - 1;
      DegU2 = Length[DeleteCases[U2, a_?Positive]] - 1;
      Q = 
        Flatten[Outer[
            List[Join[{v1}, Part[Union[U1], {#1, #1 + 1}]], 
                Join[{v2}, Part[Union[U2], {#2, #2 + 1}]]] &, 
            Table[i, {i, 1, Length[Union[U1]] - 1}], 
            Table[j, {j, 1, Length[Union[U2]] - 1}]], 1];
      P = 
        Map[Join[
              Flatten[Simplify[Map[SurfaceBasis2.# &, SurfaceBasis1.pts]], 
                  1][[#]], Q[[#]]] &, 
          Table[i, {i, 
              1, (Length[Transpose[SurfaceBasis1]] - 
                    DegU1) (Length[Transpose[SurfaceBasis2]] - DegU2)}]];
      Return[P]];

Options[PlotBSplineSurface] = {DisplayFunction -> $DisplayFunction};

PlotBSplineSurface[pts : {{{_, _, _} ..} ..}, {U1_, U2_}, {v1_, v2_}, 
      Opt___] := Module[{},
      BSS = BSplineSurface[pts, {U1, U2}, {v1, v2}];
      SurfaceGraph = 
        Map[ParametricPlot3D[Take[#, 3], Evaluate[#[[4]]], Evaluate[#[[5]]], 
              DisplayFunction -> Identity] &, BSS];
      SurfaceNet = Graphics3D[Map[Line[#] &, Join[pts, Transpose[pts]]]];
      SurfacePoints = 
        Graphics3D[Map[{PointSize[0.015], Point[#]} &, Flatten[pts, 1]]];
      Show[SurfaceGraph, SurfaceNet, SurfacePoints, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotBSplineSurface]), Opt]];

NURBSCurve[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, U_, v_] := Module[{},
      CurveBasis = NBasis[U, v];
      If[Length[First[pt]] == 3, 
        ptw = pt /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ} :> 
                If[c == 0, {a, b, c}, {a c, b c, c}]}, 
        ptw = pt /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ, d_?NumberQ} :> 
                If[d == 0, {a, b, c, d}, {a d, b d, c d, d}]}];
      P = 
        Map[Append[Simplify[CurveBasis.ptw][[#]], 
              Join[{v}, Part[Union[U], {#, # + 1}]]] &, 
          Table[i, {i, 1, Length[Union[U]] - 1}]];
      Q = Map[Append[Drop[Drop[#, -1]/Last[Drop[#, -1]], -1], Last[#]] &, P];
      Return[Q]];

Options[PlotNURBSCurve] = {DisplayFunction -> $DisplayFunction, 
      Color -> 0.6};

PlotNURBSCurveClassic[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, U_, v_, 
      Opt___] := Module[{NC},
      NC = NURBSCurve[pt, U, v];
      If[Length[pt[[1]]] == 3, Module[{},
          
          CurveGraph = 
            Map[ParametricPlot[Drop[#, -1], Evaluate[Last[#]], 
                  PlotStyle -> Hue[Color /. {Opt} /. Options[PlotNURBSCurve]],
                   DisplayFunction -> Identity] &, NC];
          
          CurvePolygon = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> True, 
              DisplayFunction -> Identity];
          
          CurvePoints = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> False, 
              PlotStyle -> PointSize[0.02], DisplayFunction -> Identity];
          
          CurveWeights = 
            Graphics[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]], Module[{},
          
          CurveGraph = 
            Map[ParametricPlot3D[
                  Join[Drop[#, -1], {Hue[
                        Color /. {Opt} /. Options[PlotNURBSCurve]]}], 
                  Evaluate[Last[#]], PlotStyle -> Hue[1], 
                  DisplayFunction -> Identity] &, NC];
          CurvePolygon = Graphics3D[Line[Map[Drop[#, -1] &, pt]]];
          
          CurvePoints = 
            Graphics3D[
              Map[{PointSize[0.02], Point[#]} &, Map[Drop[#, -1] &, pt]]];
          
          CurveWeights = 
            Graphics3D[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]
          ]];
      Show[CurveGraph, CurvePolygon, CurvePoints, CurveWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotNURBSCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

PlotNURBSCurve[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, U_, v_, Opt___] := 
    Module[{NC},
      NC = NURBSCurve[pt, U, v];
      If[Length[pt[[1]]] == 3, Module[{},
          
          CurveGraph = 
            Map[ParametricPlot[Drop[#, -1], Evaluate[Last[#]], 
                  PlotStyle -> Hue[Color /. {Opt} /. Options[PlotNURBSCurve]],
                   DisplayFunction -> Identity] &, NC];
          PN = Flatten[Position[pt, _?(#[[3]] == 0 &), 1, Heads -> False]];
          
          PolygonPoints = 
            If[Length[PN] > 0, 
              Append[Prepend[
                  Table[Append[
                      Prepend[pt[[Range[PN[[i]] + 1, PN[[i + 1]] - 1]]], 
                        pt[[PN[[i]]]] + pt[[PN[[i]] + 1]]], 
                      pt[[PN[[i + 1]] - 1]] + pt[[PN[[i + 1]]]]], {i, 1, 
                      Length[PN] - 1}], 
                  Append[pt[[Range[PN[[1]] - 1]]], 
                    pt[[PN[[1]] - 1]] + pt[[PN[[1]]]]]], 
                Prepend[pt[[Range[Last[PN] + 1, Length[pt]]]], 
                  pt[[Last[PN]]] + pt[[Last[PN] + 1]]]], {pt}];
          
          CurvePolygon = 
            Show[Map[
                ListPlot[Transpose[Drop[Transpose[#], -1]], 
                    PlotJoined -> True, DisplayFunction -> Identity] &, 
                PolygonPoints]];
          
          CurvePoints = 
            ListPlot[Map[Drop[#, -1] &, pt], PlotJoined -> False, 
              PlotStyle -> PointSize[0.02], DisplayFunction -> Identity];
          
          CurveWeights = 
            Graphics[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]], Module[{},
          
          CurveGraph = 
            Map[ParametricPlot3D[
                  Join[Drop[#, -1], {Hue[
                        Color /. {Opt} /. Options[PlotNURBSCurve]]}], 
                  Evaluate[Last[#]], DisplayFunction -> Identity] &, NC];
          PN = Flatten[Position[pt, _?(#[[4]] == 0 &), 1, Heads -> False]];
          
          PolygonPoints = 
            If[Length[PN] > 0, 
              Append[Prepend[
                  Table[Append[
                      Prepend[pt[[Range[PN[[i]] + 1, PN[[i + 1]] - 1]]], 
                        pt[[PN[[i]]]] + pt[[PN[[i]] + 1]]], 
                      pt[[PN[[i + 1]] - 1]] + pt[[PN[[i + 1]]]]], {i, 1, 
                      Length[PN] - 1}], 
                  Append[pt[[Range[PN[[1]] - 1]]], 
                    pt[[PN[[1]] - 1]] + pt[[PN[[1]]]]]], 
                Prepend[pt[[Range[Last[PN] + 1, Length[pt]]]], 
                  pt[[Last[PN]]] + pt[[Last[PN] + 1]]]], {pt}];
          
          CurvePolygon = 
            Show[Map[
                Graphics3D[Line[Transpose[Drop[Transpose[#], -1]]], 
                    DisplayFunction -> Identity] &, PolygonPoints]];
          
          CurvePoints = 
            Graphics3D[
              Map[{PointSize[0.02], Point[#]} &, Map[Drop[#, -1] &, pt]]];
          
          CurveWeights = 
            Graphics3D[
              Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1], {0, 0}]} &, 
                pt]]
          ]];
      Show[CurveGraph, CurvePolygon, CurvePoints, CurveWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotNURBSCurve]), 
        Apply[Sequence, 
          DeleteCases[{Opt}, _?(First[# /. {Rule -> List}] == Color &)]]]];

NURBSSurface[pts : {{{_, _, _, _} ..} ..}, {U1_, U2_}, {v1_, v2_}] := 
    Module[{},
      SurfaceBasis1 = NBasis[U1, v1];
      SurfaceBasis2 = NBasis[U2, v2];
      DegU1 = Length[DeleteCases[U1, a_?Positive]] - 1;
      DegU2 = Length[DeleteCases[U2, a_?Positive]] - 1;
      ptsw = 
        pts /. {{a_?NumberQ, b_?NumberQ, c_?NumberQ, d_?NumberQ} :> 
              If[d == 0, {a, b, c, d}, {a d, b d, c d, d}]};
      Q = 
        Flatten[Outer[
            List[Join[{v1}, Part[Union[U1], {#1, #1 + 1}]], 
                Join[{v2}, Part[Union[U2], {#2, #2 + 1}]]] &, 
            Table[i, {i, 1, Length[Union[U1]] - 1}], 
            Table[j, {j, 1, Length[Union[U2]] - 1}]], 1];
      P = 
        Map[Join[
              Flatten[Simplify[Map[SurfaceBasis2.# &, SurfaceBasis1.ptsw]], 
                  1][[#]], Q[[#]]] &, Table[i, {i, 1, Length[Q]}]];
      R = 
        Map[Join[Drop[Drop[#, -2]/Last[Drop[#, -2]], -1], Take[#, -2]] &, P];
      Return[R]];

Options[PlotNURBSSurface] = {DisplayFunction -> $DisplayFunction};

PlotNURBSSurface[pts : {{{_, _, _, _} ..} ..}, {U1_, U2_}, {v1_, v2_}, 
      Opt___] := Module[{},
      NS = NURBSSurface[pts, {U1, U2}, {v1, v2}];
      SurfaceGraph = 
        Map[ParametricPlot3D[Take[#, 3], Evaluate[#[[4]]], Evaluate[#[[5]]], 
              DisplayFunction -> Identity] &, NS];
      SurfaceNet = 
        Graphics3D[
          Map[Line[Map[Drop[#, -1] &, #]] &, Join[pts, Transpose[pts]]]];
      SurfacePoints = 
        Graphics3D[
          Map[{PointSize[0.015], Point[Drop[#, -1]]} &, Flatten[pts, 1]]];
      SurfaceWeights = 
        Graphics3D[
          Map[{RGBColor[0, 1, 1], Text[Last[#], Drop[#, -1]]} &, 
            Flatten[pts, 1]]];
      Show[SurfaceGraph, SurfaceNet, SurfacePoints, SurfaceWeights, 
        DisplayFunction -> (DisplayFunction /. {Opt} /. 
              Options[PlotNURBSSurface]), Opt]];

Rotation2D[pt : {{_, _} ..} | {{_, _, _} ..}, alfa_] := 
    If[Length[First[pt]] == 2, 
      Map[#.{{Cos[alfa], Sin[alfa]}, {-Sin[alfa], Cos[alfa]}} &, pt], 
      Map[#.{{Cos[alfa], Sin[alfa], 0}, {-Sin[alfa], Cos[alfa], 0}, {0, 0, 
                1}} &, pt]];

Reflection2D[pt : {{_, _} ..} | {{_, _, _} ..}, coef : {_, _, _}] := 
    Module[{},
      {l1, l2} = -2 coef[[{1, 2}]]/Apply[Plus, coef^2];
      If[Length[First[pt]] == 2,
        Module[{},
          
          T = {{1 + l1 coef[[1]], l2 coef[[1]]}, {l1 coef[[2]], 
                1 + l2 coef[[2]]}};
          b = {l1 coef[[3]], l2 coef[[3]]};
          Map[(#.T + b) &, pt]],
        Module[{},
          
          T = {{1 + l1 coef[[1]], l2 coef[[1]], 0}, {l2 coef[[2]], 
                1 + l2 coef[[2]], 0}, {l1 coef[[3]], l2 coef[[3]], 1}};
          Map[#.T &, pt]]]];

Reflection3D[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, coef : {_, _, _, _}] := 
    Module[{},
      {l1, l2, l3} = -2 coef[[{1, 2, 3}]]/Apply[Plus, coef^2];
      If[Length[First[pt]] == 3,
        Module[{},
          
          T = {{1 + l1 coef[[1]], l2 coef[[1]], l3 coef[[1]]}, {l1 coef[[2]], 
                1 + l2 coef[[2]], l3 coef[[2]]}, {l1 coef[[3]], l2 coef[[3]], 
                1 + l3 coef[[3]]}};
          b = {l1 coef[[4]], l2 coef[[4]], l3 coef[[4]]};
          Map[(#.T + b) &, pt]],
        Module[{},
          
          T = {{1 + l1 coef[[1]], l2 coef[[1]], l3 coef[[1]], 
                0}, {l1 coef[[2]], 1 + l2 coef[[2]], l3 coef[[2]], 
                0}, {l1 coef[[3]], l2 coef[[3]], 1 + l3 coef[[3]], 
                0}, {l1 coef[[4]], l2 coef[[4]], l3 coef[[4]], 1}};
          Map[#.T &, pt]]]];

Rotation3D[pt : {{_, _, _} ..} | {{_, _, _, _} ..}, {u_, A_}, alfa_] := 
    Module[{},
      u1 = u/Norm[u];
      Ac = {{0, -u1[[3]], u1[[2]]}, {u1[[3]], 0, -u1[[1]]}, {-u1[[2]], 
            u1[[1]], 0}};
      T = 
        Outer[Times, u1, u1] + 
          Cos[alfa] (IdentityMatrix[3] - Outer[Times, u1, u1]) + 
          Sin[alfa] Ac;
      b = (IdentityMatrix[3] - T).A;
      If[Length[First[pt]] == 3, Map[(#.Transpose[T] + b) &, pt], 
        Map[Join[Take[#, 3].Transpose[T] + b, {Last[#]}] &, pt]]];

RevolutionSurface[ptrevpoints : {{_, _, _, _} ..}, 
      VV_, {v1_, v2_}, {rp_, AP_}] := Module[{},
      ControlPoints = Transpose[Map[Module[{},
                primka = AP + rp t;
                rhonew = rp.{x, y, z} - rp.Take[#, 3];
                
                Snew = Flatten[
                    primka /. 
                      Solve[(rhonew /. {x -> primka[[1]], y -> primka[[2]], 
                                z -> primka[[3]]}) == 0, t]];
                dnew = Norm[Snew - Take[#, 3]];
                
                Tnew = {{1, 0, 0, 1}, {1, 1, 0, 1/2}, {-1, 1, 0, 1/2}, {-1, 0,
                         0, 1}, {-1, -1, 0, 1/2}, {1, -1, 0, 1/2}, {1, 0, 0, 
                        1}} /. {a1_?NumberQ, b1_?NumberQ, c1_?NumberQ, 
                        e1_?NumberQ} -> {a1 dnew + Snew[[1]], 
                        b1 dnew + Snew[[2]], c1 dnew + Snew[[3]], e1 #[[4]]};
                
                directionnew = 
                  Cross[Take[First[Tnew], 3] - Snew, Take[#, 3] - Snew];
                
                If[Norm[directionnew] == 0, Tnew, 
                  Rotation3D[Tnew, {directionnew, Snew}, 
                    ArcCos[((Take[First[Tnew], 3] - Snew).(Take[#, 3] - 
                                Snew))/(Norm[
                              Take[First[Tnew], 3] - Snew] Norm[
                              Take[#, 3] - Snew])]]]] &, ptrevpoints]];
      UU = {0, 0, 0, 1/4, 1/2, 1/2, 3/4, 1, 1, 1};
      NURBSSurface[ControlPoints, {UU, VV}, {v1, v2}]];

PlotRevolutionSurface[ptrevpoints : {{_, _, _, _} ..}, 
      VV_, {v1_, v2_}, {rp_, AP_}] := Module[{},
      ControlPoints2 = Transpose[Map[Module[{},
                primka2 = AP + rp t;
                rhonew2 = rp.{x, y, z} - rp.Take[#, 3];
                
                Snew2 = 
                  Flatten[
                    primka2 /. 
                      Solve[(rhonew2 /. {x -> primka2[[1]], y -> primka2[[2]],
                                 z -> primka2[[3]]}) == 0, t]];
                dnew2 = Norm[Snew2 - Take[#, 3]];
                
                Tnew2 = {{1, 0, 0, 1}, {1, 1, 0, 1/2}, {-1, 1, 0, 1/2}, {-1, 
                        0, 0, 1}, {-1, -1, 0, 1/2}, {1, -1, 0, 1/2}, {1, 0, 0,
                         1}} /. {a2_?NumberQ, b2_?NumberQ, c2_?NumberQ, 
                        e2_?NumberQ} -> {a2 dnew2 + Snew2[[1]], 
                        b2 dnew2 + Snew2[[2]], c2 dnew2 + Snew2[[3]], 
                        e2 #[[4]]};
                
                directionnew2 = 
                  Cross[Take[First[Tnew2], 3] - Snew2, Take[#, 3] - Snew2];
                
                If[Norm[directionnew2] == 0, Tnew2, 
                  Rotation3D[Tnew2, {directionnew2, Snew2}, 
                    ArcCos[((Take[First[Tnew2], 3] - Snew2).(Take[#, 3] - 
                                Snew2))/(Norm[
                              Take[First[Tnew2], 3] - Snew2] Norm[
                              Take[#, 3] - Snew2])]]]] &, ptrevpoints]];
      UU = {0, 0, 0, 1/4, 1/2, 1/2, 3/4, 1, 1, 1};
      PlotNURBSSurface[ControlPoints2, {UU, VV}, {v1, v2}]];

End[]

EndPackage[]

Print["\n\nPackage CAGD.m,  Version 1.1,  July 2005\n\n\n
User-accessible functions:\n\n
{ NormalizeVectors, ChordalParametrization, Ferguson, CubicSpline, PlotCubicSpline, QuadraticSpline, PlotQuadraticSpline \n\nBernstein, BezierCurve, PlotBezierCurve, RatBezierCurve, PlotRatBezierCurve, BezierSurface, PlotBezierSurface, RatBezierSurface, PlotRatBezierSurface, NBasis, PlotNBasis, BSplineCurve, PlotBSplineCurve, BSplineSurface, PlotBSplineSurface, NURBSCurve, PlotNURBSCurve, NURBSSurface, PlotNURBSSurface, \n\n Rotation2D, Reflection2D, Rotation3D, Reflection3D, \n\n RevolutionSurface, PlotRevolutionSurface }\n\n
"]
